;; Generated package description from names.el  -*- no-byte-compile: t -*-
(define-package "names" "20151201.0.0.201809201516" "Namespaces for emacs-lisp. Avoid name clobbering without hiding symbols." '((emacs "24.1") (cl-lib "0.5") (nadvice "0.3")) :keywords '("extensions" "lisp") :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :url "https://github.com/Malabarba/names")
